#  Copyright (c) 2020. Lena "Teekeks" During <info@teawork.de>
"""
Type Definitions
----------------"""
from dataclasses import dataclass
from enum import Enum
from typing_extensions import TypedDict
from enum_tools.documentation import document_enum

__all__ = ['AnalyticsReportType', 'AuthScope', 'ModerationEventType', 'TimePeriod', 'SortMethod', 'HypeTrainContributionMethod',
           'VideoType', 'AuthType', 'StatusCode', 'CustomRewardRedemptionStatus', 'SortOrder',
           'BlockSourceContext', 'BlockReason', 'EntitlementFulfillmentStatus', 'PollStatus', 'PredictionStatus', 'AutoModAction',
           'AutoModCheckEntry', 'DropsEntitlementFulfillmentStatus', 'ChatEvent', 'ChatRoom',
           'TwitchAPIException', 'InvalidRefreshTokenException', 'InvalidTokenException', 'NotFoundException', 'TwitchAuthorizationException',
           'UnauthorizedException', 'MissingScopeException', 'TwitchBackendException', 'MissingAppSecretException',
           'EventSubSubscriptionTimeout', 'EventSubSubscriptionConflict', 'EventSubSubscriptionError', 'DeprecatedError', 'TwitchResourceNotFound',
           'ForbiddenError']


class AnalyticsReportType(Enum):
    """Enum of all Analytics report types
    """
    V1 = 'overview_v1'
    V2 = 'overview_v2'

@document_enum
class AuthScope(Enum):
    """Enum of Authentication scopes"""
    ANALYTICS_READ_EXTENSION = 'analytics:read:extensions'
    """View analytics data for the Twitch Extensions owned by the authenticated account.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_game_analytics()`
    """
    ANALYTICS_READ_GAMES = 'analytics:read:games'
    """View analytics data for the games owned by the authenticated account.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_game_analytics()`
    """
    BITS_READ = 'bits:read'
    """View Bits information for a channel.
             
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_bits_leaderboard()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_cheer()`
    """
    CHANNEL_READ_SUBSCRIPTIONS = 'channel:read:subscriptions'
    """View a list of all subscribers to a channel and check if a user is subscribed to a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_broadcaster_subscriptions()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_subscribe()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_subscription_end()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_subscription_gift()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_subscription_message()` |br|
    """
    CHANNEL_READ_STREAM_KEY = 'channel:read:stream_key'
    """View an authorized user’s stream key.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_stream_key()` |br|
    """
    CHANNEL_EDIT_COMMERCIAL = 'channel:edit:commercial'
    """Run commercials on a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.start_commercial()`
    """
    CHANNEL_READ_HYPE_TRAIN = 'channel:read:hype_train'
    """View Hype Train information for a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_hype_train_events()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_hype_train_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_hype_train_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_hype_train_end()` |br|
    """
    CHANNEL_MANAGE_BROADCAST = 'channel:manage:broadcast'
    """Manage a channel’s broadcast configuration, including updating channel configuration and managing stream markers and stream tags.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.modify_channel_information()` |br|
    :const:`~twitchAPI.twitch.Twitch.create_stream_marker()`
    """
    CHANNEL_READ_REDEMPTIONS = 'channel:read:redemptions'
    """View Channel Points custom rewards and their redemptions on a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_custom_reward()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_custom_reward_redemption()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_automatic_reward_redemption_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_update()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_remove()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_redemption_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_redemption_update()` |br|
    """
    CHANNEL_MANAGE_REDEMPTIONS = 'channel:manage:redemptions'
    """Manage Channel Points custom rewards and their redemptions on a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_custom_reward()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_custom_reward_redemption()` |br|
    :const:`~twitchAPI.twitch.Twitch.create_custom_reward()` |br|
    :const:`~twitchAPI.twitch.Twitch.delete_custom_reward()` |br|
    :const:`~twitchAPI.twitch.Twitch.update_custom_reward()` |br|
    :const:`~twitchAPI.twitch.Twitch.update_redemption_status()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_automatic_reward_redemption_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_update()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_remove()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_redemption_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_points_custom_reward_redemption_update()` |br|
    """
    CHANNEL_READ_CHARITY = 'channel:read:charity'
    """Read charity campaign details and user donations on your channel.
           
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_charity_campaign()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_charity_donations()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_charity_campaign_donate()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_charity_campaign_start()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_charity_campaign_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_charity_campaign_stop()` |br|
    """
    CLIPS_EDIT = 'clips:edit'
    """Manage Clips for a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.create_clip()` |br|
    """
    USER_EDIT = 'user:edit'
    """Manage a user object.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_user()` |br|
    """
    USER_EDIT_BROADCAST = 'user:edit:broadcast'
    """View and edit a user’s broadcasting configuration, including Extension configurations.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_extensions()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_active_extensions()` |br|
    :const:`~twitchAPI.twitch.Twitch.update_user_extensions()` |br|
    """
    USER_READ_BROADCAST = 'user:read:broadcast'
    """View a user’s broadcasting configuration, including Extension configurations.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_stream_markers()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_extensions()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_active_extensions()` |br|
    """
    USER_READ_EMAIL = 'user:read:email'
    """View a user’s email address.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_users()` (optional) |br|
    :const:`~twitchAPI.twitch.Twitch.update_user()` (optional) |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_user_update()` (optional) |br|
    """
    USER_EDIT_FOLLOWS = 'user:edit:follows'
    CHANNEL_MODERATE = 'channel:moderate'
    CHAT_EDIT = 'chat:edit'
    """Send chat messages to a chatroom using an IRC connection."""
    CHAT_READ = 'chat:read'
    """View chat messages sent in a chatroom using an IRC connection."""
    WHISPERS_READ = 'whispers:read'
    """Receive whisper messages for your user using PubSub."""
    WHISPERS_EDIT = 'whispers:edit'
    MODERATION_READ = 'moderation:read'
    """
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.check_automod_status()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_banned_users()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_moderators()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderator_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderator_remove()` |br|
    """
    CHANNEL_SUBSCRIPTIONS = 'channel_subscriptions'
    CHANNEL_READ_EDITORS = 'channel:read:editors'
    """View a list of users with the editor role for a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_channel_editors()`
    """
    CHANNEL_MANAGE_VIDEOS = 'channel:manage:videos'
    """Manage a channel’s videos, including deleting videos.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.delete_videos()` |br|
    """
    USER_READ_BLOCKED_USERS = 'user:read:blocked_users'
    """View the block list of a user.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_block_list()` |br|
    """
    USER_MANAGE_BLOCKED_USERS = 'user:manage:blocked_users'
    """Manage the block list of a user.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.block_user()` |br|
    :const:`~twitchAPI.twitch.Twitch.unblock_user()` |br|
    """
    USER_READ_SUBSCRIPTIONS = 'user:read:subscriptions'
    """View if an authorized user is subscribed to specific channels.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.check_user_subscription()` |br|
    """
    USER_READ_FOLLOWS = 'user:read:follows'
    """View the list of channels a user follows.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_followed_channels()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_followed_streams()` |br|
    """
    CHANNEL_READ_GOALS = 'channel:read:goals'
    """View Creator Goals for a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_creator_goals()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_goal_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_goal_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_goal_end()` |br|
    """
    CHANNEL_READ_POLLS = 'channel:read:polls'
    """View a channel’s polls.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_polls()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_end()` |br|
    """
    CHANNEL_MANAGE_POLLS = 'channel:manage:polls'
    """Manage a channel’s polls.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_polls()` |br|
    :const:`~twitchAPI.twitch.Twitch.create_poll()` |br|
    :const:`~twitchAPI.twitch.Twitch.end_poll()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_poll_end()` |br|
    """
    CHANNEL_READ_PREDICTIONS = 'channel:read:predictions'
    """View a channel’s Channel Points Predictions.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_predictions()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_lock()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_end()` |br|
    """
    CHANNEL_MANAGE_PREDICTIONS = 'channel:manage:predictions'
    """Manage of channel’s Channel Points Predictions
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_predictions()` |br|
    :const:`~twitchAPI.twitch.Twitch.create_prediction()` |br|
    :const:`~twitchAPI.twitch.Twitch.end_prediction()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_progress()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_lock()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_prediction_end()` |br|
    """
    MODERATOR_MANAGE_AUTOMOD = 'moderator:manage:automod'
    """Manage messages held for review by AutoMod in channels where you are a moderator.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.manage_held_automod_message()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_automod_message_hold()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_automod_message_update()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_automod_terms_update()` |br|
    """
    CHANNEL_MANAGE_SCHEDULE = 'channel:manage:schedule'
    """Manage a channel’s stream schedule.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_channel_stream_schedule()` |br|
    :const:`~twitchAPI.twitch.Twitch.create_channel_stream_schedule_segment()` |br|
    :const:`~twitchAPI.twitch.Twitch.update_channel_stream_schedule_segment()` |br|
    :const:`~twitchAPI.twitch.Twitch.delete_channel_stream_schedule_segment()` |br|
    """
    MODERATOR_MANAGE_CHAT_SETTINGS = 'moderator:manage:chat_settings'
    """Manage a broadcaster’s chat room settings.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_chat_settings()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_CHAT_SETTINGS = 'moderator:read:chat_settings'
    """View a broadcaster’s chat room settings.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_chat_settings()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|"""
    MODERATOR_MANAGE_BANNED_USERS = 'moderator:manage:banned_users'
    """Ban and unban users.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_banned_users()` |br|
    :const:`~twitchAPI.twitch.Twitch.ban_user()` |br|
    :const:`~twitchAPI.twitch.Twitch.unban_user()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_BANNED_USERS = 'moderator:read:banned_users'
    """Read banned users.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_BLOCKED_TERMS = 'moderator:read:blocked_terms'
    """View a broadcaster’s list of blocked terms.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_blocked_terms()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_MANAGE_BLOCKED_TERMS = 'moderator:manage:blocked_terms'
    """Manage a broadcaster’s list of blocked terms.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_blocked_terms()` |br|
    :const:`~twitchAPI.twitch.Twitch.add_blocked_term()` |br|
    :const:`~twitchAPI.twitch.Twitch.remove_blocked_term()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    CHANNEL_MANAGE_RAIDS = 'channel:manage:raids'
    """Manage a channel raiding another channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.start_raid()` |br|
    :const:`~twitchAPI.twitch.Twitch.cancel_raid()` |br|
    """
    MODERATOR_MANAGE_ANNOUNCEMENTS = 'moderator:manage:announcements'
    """Send announcements in channels where you have the moderator role.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_chat_announcement()` |br|
    """
    MODERATOR_MANAGE_CHAT_MESSAGES = 'moderator:manage:chat_messages'
    """Delete chat messages in channels where you have the moderator role.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.delete_chat_message()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_CHAT_MESSAGES = 'moderator:read:chat_messages'
    """Read deleted chat messages in channels where you have the moderator role.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_WARNINGS = 'moderator:read:warnings'
    """Read warnings in channels where you have the moderator role.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_warning_acknowledge()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_warning_send()` |br|
    """
    MODERATOR_MANAGE_WARNINGS = 'moderator:manage:warnings'
    """Warn users in channels where you have the moderator role.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.warn_chat_user()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_warning_acknowledge()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_warning_send()` |br|
    """
    USER_MANAGE_CHAT_COLOR = 'user:manage:chat_color'
    """Update the color used for the user’s name in chat.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_user_chat_color()` |br|
    """
    CHANNEL_MANAGE_MODERATORS = 'channel:manage:moderators'
    """Add or remove the moderator role from users in your channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.add_channel_moderator()` |br|
    :const:`~twitchAPI.twitch.Twitch.remove_channel_moderator()` |br|
    :const:`~twitchAPI.twitch.Twitch.get_moderators()` |br|
    """
    CHANNEL_READ_VIPS = 'channel:read:vips'
    """Read the list of VIPs in your channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_vips()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_vip_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_vip_remove()` |br|
    """
    MODERATOR_READ_MODERATORS = 'moderator:read:moderators'
    """Read the list of channels you are moderator in.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_VIPS = 'moderator:read:vips'
    CHANNEL_MANAGE_VIPS = 'channel:manage:vips'
    """Add or remove the VIP role from users in your channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_vips()` |br|
    :const:`~twitchAPI.twitch.Twitch.add_channel_vip()` |br|
    :const:`~twitchAPI.twitch.Twitch.remove_channel_vip()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_vip_add()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_vip_remove()` |br|
    """
    USER_READ_WHISPERS = 'user:read:whispers'
    """Receive whispers sent to your user.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_user_whisper_message()` |br|
    """
    USER_MANAGE_WHISPERS = 'user:manage:whispers'
    """Receive whispers sent to your user, and send whispers on your user’s behalf.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_whisper()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_user_whisper_message()` |br|
    """
    MODERATOR_READ_CHATTERS = 'moderator:read:chatters'
    """View the chatters in a broadcaster’s chat room.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_chatters()` |br|
    """
    MODERATOR_READ_SHIELD_MODE = 'moderator:read:shield_mode'
    """View a broadcaster’s Shield Mode status.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_shield_mode_status()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shield_mode_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shield_mode_end()` |br|
    """
    MODERATOR_MANAGE_SHIELD_MODE = 'moderator:manage:shield_mode'
    """Manage a broadcaster’s Shield Mode status.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_shield_mode_status()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shield_mode_begin()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shield_mode_end()` |br|
    """
    MODERATOR_READ_AUTOMOD_SETTINGS = 'moderator:read:automod_settings'
    """View a broadcaster’s AutoMod settings.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_automod_settings()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_automod_settings_update()` |br|
    """
    MODERATOR_MANAGE_AUTOMOD_SETTINGS = 'moderator:manage:automod_settings'
    """Manage a broadcaster’s AutoMod settings.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.update_automod_settings()` |br|
    """
    MODERATOR_READ_FOLLOWERS = 'moderator:read:followers'
    """Read the followers of a broadcaster.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_channel_followers()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_follow_v2()` |br|
    """
    MODERATOR_MANAGE_SHOUTOUTS = 'moderator:manage:shoutouts'
    """Manage a broadcaster’s shoutouts.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_a_shoutout()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shoutout_create()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shoutout_receive()` |br|
    """
    MODERATOR_READ_SHOUTOUTS = 'moderator:read:shoutouts'
    """View a broadcaster’s shoutouts.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shoutout_create()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_shoutout_receive()` |br|
    """
    CHANNEL_BOT = 'channel:bot'
    """Joins your channel’s chatroom as a bot user, and perform chat-related actions as that user.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_chat_message()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_clear_user_messages()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message_delete()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_notification()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_settings_update()` 
    """
    USER_BOT = 'user:bot'
    """Join a specified chat channel as your user and appear as a bot, and perform chat-related actions as your user.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_chat_message()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_clear()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_clear_user_messages()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message_delete()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_notification()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_settings_update()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_user_message_hold()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_user_message_update()` |br|
    """
    USER_READ_CHAT = 'user:read:chat'
    """Receive chatroom messages and informational notifications relating to a channel’s chatroom.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_clear()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_clear_user_messages()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_message_delete()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_notification()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_settings_update()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_user_message_hold()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_chat_user_message_update()` |br|
    """
    CHANNEL_READ_ADS = 'channel:read:ads'
    """Read the ads schedule and details on your channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_ad_schedule()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_ad_break_begin()`
    """
    CHANNEL_MANAGE_ADS = 'channel:manage:ads'
    """Manage ads schedule on a channel.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_ad_schedule()`
    """
    USER_WRITE_CHAT = 'user:write:chat'
    """Send chat messages to a chatroom.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.send_chat_message()` |br|
    """
    USER_READ_MODERATED_CHANNELS = 'user:read:moderated_channels'
    """Read the list of channels you have moderator privileges in.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_moderated_channels()` |br|
    """
    USER_READ_EMOTES = 'user:read:emotes'
    """View emotes available to a user.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_user_emotes()` |br|
    """
    MODERATOR_READ_UNBAN_REQUESTS = 'moderator:read:unban_requests'
    """View a broadcaster’s unban requests.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.get_unban_requests()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_unban_request_create()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_unban_request_resolve()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_MANAGE_UNBAN_REQUESTS = 'moderator:manage:unban_requests'
    """Manage a broadcaster’s unban requests.
    
    **API** |br|
    :const:`~twitchAPI.twitch.Twitch.resolve_unban_requests()` |br|
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_unban_request_create()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_unban_request_resolve()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_moderate()` |br|
    """
    MODERATOR_READ_SUSPICIOUS_USERS = 'moderator:read:suspicious_users'
    """Read chat messages from suspicious users and see users flagged as suspicious in channels where you have the moderator role.
    
    **EventSub** |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_suspicious_user_message()` |br|
    :const:`~twitchAPI.eventsub.base.EventSubBase.listen_channel_suspicious_user_update()` |br|
    """

class ModerationEventType(Enum):
    """Enum of moderation event types
    """
    BAN = 'moderation.user.ban'
    UNBAN = 'moderation.user.unban'
    UNKNOWN = ''


class TimePeriod(Enum):
    """Enum of valid Time periods
    """
    ALL = 'all'
    DAY = 'day'
    WEEK = 'week'
    MONTH = 'month'
    YEAR = 'year'


class SortMethod(Enum):
    """Enum of valid sort methods
    """
    TIME = 'time'
    TRENDING = 'trending'
    VIEWS = 'views'


class HypeTrainContributionMethod(Enum):
    """Enum of valid Hype Train contribution types
    """

    BITS = 'BITS'
    SUBS = 'SUBS'
    OTHER = 'OTHER'
    UNKNOWN = ''


class VideoType(Enum):
    """Enum of valid video types
    """
    ALL = 'all'
    UPLOAD = 'upload'
    ARCHIVE = 'archive'
    HIGHLIGHT = 'highlight'
    UNKNOWN = ''


class AuthType(Enum):
    """Type of authentication required. Only internal use
    """
    NONE = 0
    USER = 1
    APP = 2
    EITHER = 3


class StatusCode(Enum):
    """Enum Code Status, see https://dev.twitch.tv/docs/api/reference#get-code-status for more documentation
    """
    SUCCESSFULLY_REDEEMED = 'SUCCESSFULLY_REDEEMED'
    ALREADY_CLAIMED = 'ALREADY_CLAIMED'
    EXPIRED = 'EXPIRED'
    USER_NOT_ELIGIBLE = 'USER_NOT_ELIGIBLE'
    NOT_FOUND = 'NOT_FOUND'
    INACTIVE = 'INACTIVE'
    UNUSED = 'UNUSED'
    INCORRECT_FORMAT = 'INCORRECT_FORMAT'
    INTERNAL_ERROR = 'INTERNAL_ERROR'
    UNKNOWN_VALUE = ''


class CustomRewardRedemptionStatus(Enum):
    """
    """
    UNFULFILLED = 'UNFULFILLED'
    FULFILLED = 'FULFILLED'
    CANCELED = 'CANCELED'


class SortOrder(Enum):
    """
    """
    OLDEST = 'OLDEST'
    NEWEST = 'NEWEST'


class BlockSourceContext(Enum):
    """
    """
    CHAT = 'chat'
    WHISPER = 'whisper'


class BlockReason(Enum):
    """
    """
    SPAM = 'spam'
    HARASSMENT = 'harassment'
    OTHER = 'other'


class EntitlementFulfillmentStatus(Enum):
    """
    """
    CLAIMED = 'CLAIMED'
    FULFILLED = 'FULFILLED'


class PollStatus(Enum):
    """
    """
    ACTIVE = 'ACTIVE'
    COMPLETED = 'COMPLETED'
    MODERATED = 'MODERATED'
    INVALID = 'INVALID'
    TERMINATED = 'TERMINATED'
    ARCHIVED = 'ARCHIVED'


class PredictionStatus(Enum):
    """
    """
    ACTIVE = 'ACTIVE'
    RESOLVED = 'RESOLVED'
    CANCELED = 'CANCELED'
    LOCKED = 'LOCKED'


class AutoModAction(Enum):
    """
    """
    ALLOW = 'ALLOW'
    DENY = 'DENY'


class DropsEntitlementFulfillmentStatus(Enum):
    """
    """
    CLAIMED = 'CLAIMED'
    FULFILLED = 'FULFILLED'


class AutoModCheckEntry(TypedDict):
    msg_id: str
    """Developer-generated identifier for mapping messages to results."""
    msg_text: str
    """Message text"""

# CHAT

@document_enum
class ChatEvent(Enum):
    """Represents the possible events to listen for using :const:`~twitchAPI.chat.Chat.register_event()`"""
    READY = 'ready'
    """Triggered when the bot is started up and ready"""
    MESSAGE = 'message'
    """Triggered when someone wrote a message in a chat channel"""
    SUB = 'sub'
    """Triggered when someone subscribed to a channel"""
    RAID = 'raid'
    """Triggered when a channel gets raided"""
    ROOM_STATE_CHANGE = 'room_state_change'
    """Triggered when a chat channel is changed (e.g. sub only mode was enabled)"""
    JOIN = 'join'
    """Triggered when someone other than the bot joins a chat channel"""
    JOINED = 'joined'
    """Triggered when the bot joins a chat channel"""
    LEFT = 'left'
    """Triggered when the bot leaves a chat channel"""
    USER_LEFT = 'user_left'
    """Triggered when a user leaves a chat channel"""
    MESSAGE_DELETE = 'message_delete'
    """Triggered when a message gets deleted from a channel"""
    CHAT_CLEARED = 'chat_cleared'
    """Triggered when a user was banned, timed out or all messaged from a user where deleted"""
    WHISPER = 'whisper'
    """Triggered when someone whispers to your bot. NOTE: You need the :const:`~twitchAPI.type.AuthScope.WHISPERS_READ` Auth Scope
    to get this Event."""
    NOTICE = 'notice'
    """Triggered on server notice"""


@dataclass
class ChatRoom:
    name: str
    is_emote_only: bool
    is_subs_only: bool
    is_followers_only: bool
    is_unique_only: bool
    follower_only_delay: int
    room_id: str
    slow: int


# EXCEPTIONS


class TwitchAPIException(Exception):
    """Base Twitch API Exception"""
    pass


class InvalidRefreshTokenException(TwitchAPIException):
    """used User Refresh Token is invalid"""
    pass


class InvalidTokenException(TwitchAPIException):
    """Used if a invalid token is set for the client"""
    pass


class NotFoundException(TwitchAPIException):
    """Resource was not found with the given parameter"""
    pass


class TwitchAuthorizationException(TwitchAPIException):
    """Exception in the Twitch Authorization"""
    pass


class UnauthorizedException(TwitchAuthorizationException):
    """Not authorized to use this"""
    pass


class MissingScopeException(TwitchAuthorizationException):
    """authorization is missing scope"""
    pass


class TwitchBackendException(TwitchAPIException):
    """when the Twitch API itself is down"""
    pass


class MissingAppSecretException(TwitchAPIException):
    """When the app secret is not set but app authorization is attempted"""
    pass


class EventSubSubscriptionTimeout(TwitchAPIException):
    """When the waiting for a confirmed EventSub subscription timed out"""
    pass


class EventSubSubscriptionConflict(TwitchAPIException):
    """When you try to subscribe to a EventSub subscription that already exists"""
    pass


class EventSubSubscriptionError(TwitchAPIException):
    """if the subscription request was invalid"""
    pass


class DeprecatedError(TwitchAPIException):
    """If something has been marked as deprecated by the Twitch API"""
    pass


class TwitchResourceNotFound(TwitchAPIException):
    """If a requested resource was not found"""
    pass


class ForbiddenError(TwitchAPIException):
    """If you are not allowed to do that"""
    pass



#  Copyright (c) 2022. Lena "Teekeks" During <info@teawork.de>
from setuptools import setup, find_packages

version = ''

with open('twitchAPI/__init__.py') as f:
    for line in f.readlines():
        if line.startswith('__version__'):
            version = line.split('= \'')[-1][:-2].strip()

if version.endswith(('a', 'b', 'rc')):
    try:
        import subprocess
        p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += out.decode('utf-8').strip()
        p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += '+g' + out.decode('utf-8').strip()
    except:
        pass

setup(
    packages=find_packages(),
    version=version,
    keywords=['twitch', 'twitch.tv', 'chat', 'bot', 'event sub', 'EventSub', 'helix', 'api'],
    install_requires=[
        'aiohttp>=3.9.3',
        'python-dateutil>=2.8.2',
        'typing_extensions',
        'enum-tools'
    ],
    package_data={'twitchAPI': ['py.typed']}
)



#  Copyright (c) 2023. Lena "Teekeks" During <info@teawork.de>
"""
EventSub Webhook
----------------

.. note:: EventSub Webhook is targeted at programs which have to subscribe to topics for multiple broadcasters.\n
    Should you only need to target a single broadcaster or are building a client side project, look at :doc:`/modules/twitchAPI.eventsub.websocket`

EventSub lets you listen for events that happen on Twitch.

The EventSub client runs in its own thread, calling the given callback function whenever an event happens.

************
Requirements
************

.. note:: Please note that Your Endpoint URL has to be HTTPS, has to run on Port 443 and requires a valid, non self signed certificate
            This most likely means, that you need a reverse proxy like nginx. You can also hand in a valid ssl context to be used in the constructor.

In the case that you don't hand in a valid ssl context to the constructor, you can specify any port you want in the constructor and handle the
bridge between this program and your public URL on port 443 via reverse proxy.\n
You can check on whether or not your webhook is publicly reachable by navigating to the URL set in `callback_url`.
You should get a 200 response with the text :code:`pyTwitchAPI eventsub`.

*******************
Listening to topics
*******************

After you started your EventSub client, you can use the :code:`listen_` prefixed functions to listen to the topics you are interested in.

Look at :ref:`eventsub-available-topics` to find the topics you are interested in.

The function you hand in as callback will be called whenever that event happens with the event data as a parameter,
the type of that parameter is also listed in the link above.

************
Code Example
************

.. code-block:: python

    from twitchAPI.twitch import Twitch
    from twitchAPI.helper import first
    from twitchAPI.eventsub.webhook import EventSubWebhook
    from twitchAPI.object.eventsub import ChannelFollowEvent
    from twitchAPI.oauth import UserAuthenticator
    from twitchAPI.type import AuthScope
    import asyncio

    TARGET_USERNAME = 'target_username_here'
    EVENTSUB_URL = 'https://url.to.your.webhook.com'
    APP_ID = 'your_app_id'
    APP_SECRET = 'your_app_secret'
    TARGET_SCOPES = [AuthScope.MODERATOR_READ_FOLLOWERS]


    async def on_follow(data: ChannelFollowEvent):
        # our event happened, lets do things with the data we got!
        print(f'{data.event.user_name} now follows {data.event.broadcaster_user_name}!')


    async def eventsub_webhook_example():
        # create the api instance and get the ID of the target user
        twitch = await Twitch(APP_ID, APP_SECRET)
        user = await first(twitch.get_users(logins=TARGET_USERNAME))

        # the user has to authenticate once using the bot with our intended scope.
        # since we do not need the resulting token after this authentication, we just discard the result we get from authenticate()
        # Please read up the UserAuthenticator documentation to get a full view of how this process works
        auth = UserAuthenticator(twitch, TARGET_SCOPES)
        await auth.authenticate()

        # basic setup, will run on port 8080 and a reverse proxy takes care of the https and certificate
        eventsub = EventSubWebhook(EVENTSUB_URL, 8080, twitch)

        # unsubscribe from all old events that might still be there
        # this will ensure we have a clean slate
        await eventsub.unsubscribe_all()
        # start the eventsub client
        eventsub.start()
        # subscribing to the desired eventsub hook for our user
        # the given function (in this example on_follow) will be called every time this event is triggered
        # the broadcaster is a moderator in their own channel by default so specifying both as the same works in this example
        await eventsub.listen_channel_follow_v2(user.id, user.id, on_follow)

        # eventsub will run in its own process
        # so lets just wait for user input before shutting it all down again
        try:
            input('press Enter to shut down...')
        finally:
            # stopping both eventsub as well as gracefully closing the connection to the API
            await eventsub.stop()
            await twitch.close()
        print('done')


    # lets run our example
    asyncio.run(eventsub_webhook_example())"""
import asyncio
import hashlib
import hmac
import threading
from functools import partial
from json import JSONDecodeError
from random import choice
from string import ascii_lowercase
from ssl import SSLContext
from time import sleep
from typing import Optional, Union, Callable, Awaitable
import datetime
from collections import deque

from aiohttp import web, ClientSession

from twitchAPI.eventsub.base import EventSubBase
from ..twitch import Twitch
from ..helper import done_task_callback
from ..type import TwitchBackendException, EventSubSubscriptionConflict, EventSubSubscriptionError, EventSubSubscriptionTimeout, \
    TwitchAuthorizationException, AuthType

__all__ = ['EventSubWebhook']


class EventSubWebhook(EventSubBase):

    def __init__(self,
                 callback_url: str,
                 port: int,
                 twitch: Twitch,
                 ssl_context: Optional[SSLContext] = None,
                 host_binding: str = '0.0.0.0',
                 subscription_url: Optional[str] = None,
                 callback_loop: Optional[asyncio.AbstractEventLoop] = None,
                 revocation_handler: Optional[Callable[[dict], Awaitable[None]]] = None,
                 message_deduplication_history_length: int = 50):
        """
        :param callback_url: The full URL of the webhook.
        :param port: the port on which this webhook should run
        :param twitch: a app authenticated instance of :const:`~twitchAPI.twitch.Twitch`
        :param ssl_context: optional ssl context to be used |default| :code:`None`
        :param host_binding: the host to bind the internal server to |default| :code:`0.0.0.0`
        :param subscription_url: Alternative subscription URL, useful for development with the twitch-cli
        :param callback_loop: The asyncio eventloop to be used for callbacks. \n
            Set this if you or a library you use cares about which asyncio event loop is running the callbacks.
            Defaults to the one used by EventSub Webhook.
        :param revocation_handler: Optional handler for when subscriptions get revoked. |default| :code:`None`
        :param message_deduplication_history_length: The amount of messages being considered for the duplicate message deduplication. |default| :code:`50`
        """
        super().__init__(twitch, 'twitchAPI.eventsub.webhook')
        self.callback_url: str = callback_url
        """The full URL of the webhook."""
        if self.callback_url[-1] == '/':
            self.callback_url = self.callback_url[:-1]
        self.secret: str = ''.join(choice(ascii_lowercase) for _ in range(20))
        """A random secret string. Set this for added security. |default| :code:`A random 20 character long string`"""
        self.wait_for_subscription_confirm: bool = True
        """Set this to false if you don't want to wait for a subscription confirm. |default| :code:`True`"""
        self.wait_for_subscription_confirm_timeout: int = 30
        """Max time in seconds to wait for a subscription confirmation. Only used if ``wait_for_subscription_confirm`` is set to True. 
            |default| :code:`30`"""

        self._port: int = port
        self.subscription_url: Optional[str] = subscription_url
        """Alternative subscription URL, useful for development with the twitch-cli"""
        if self.subscription_url is not None and self.subscription_url[-1] != '/':
            self.subscription_url += '/'
        self._callback_loop = callback_loop
        self._host: str = host_binding
        self.__running = False
        self.revokation_handler: Optional[Callable[[dict], Awaitable[None]]] = revocation_handler
        """Optional handler for when subscriptions get revoked."""
        self._startup_complete = False
        self.unsubscribe_on_stop: bool = True
        """Unsubscribe all currently active Webhooks on calling :const:`~twitchAPI.eventsub.EventSub.stop()` |default| :code:`True`"""

        self._closing = False
        self.__ssl_context: Optional[SSLContext] = ssl_context
        self.__active_webhooks = {}
        self.__hook_thread: Union['threading.Thread', None] = None
        self.__hook_loop: Union['asyncio.AbstractEventLoop', None] = None
        self.__hook_runner: Union['web.AppRunner', None] = None
        self._task_callback = partial(done_task_callback, self.logger)
        if not self.callback_url.startswith('https'):
            raise RuntimeError('HTTPS is required for authenticated webhook.\n'
                               + 'Either use non authenticated webhook or use a HTTPS proxy!')
        self._msg_id_history: deque = deque(maxlen=message_deduplication_history_length)

    async def _unsubscribe_hook(self, topic_id: str) -> bool:
        return True

    def __build_runner(self):
        hook_app = web.Application()
        hook_app.add_routes([web.post('/callback', self.__handle_callback),
                             web.get('/', self.__handle_default)])
        return web.AppRunner(hook_app)

    def __run_hook(self, runner: 'web.AppRunner'):
        self.__hook_runner = runner
        self.__hook_loop = asyncio.new_event_loop()
        if self._callback_loop is None:
            self._callback_loop = self.__hook_loop
        asyncio.set_event_loop(self.__hook_loop)
        self.__hook_loop.run_until_complete(runner.setup())
        site = web.TCPSite(runner, str(self._host), self._port, ssl_context=self.__ssl_context)
        self.__hook_loop.run_until_complete(site.start())
        self.logger.info('started twitch API event sub on port ' + str(self._port))
        self._startup_complete = True
        self.__hook_loop.run_until_complete(self._keep_loop_alive())

    async def _keep_loop_alive(self):
        while not self._closing:
            await asyncio.sleep(0.1)

    def start(self):
        """Starts the EventSub client

        :rtype: None
        :raises RuntimeError: if EventSub is already running
        """
        if self.__running:
            raise RuntimeError('already started')
        self.__hook_thread = threading.Thread(target=self.__run_hook, args=(self.__build_runner(),))
        self.__running = True
        self._startup_complete = False
        self._closing = False
        self.__hook_thread.start()
        while not self._startup_complete:
            sleep(0.1)

    async def stop(self):
        """Stops the EventSub client

        This also unsubscribes from all known subscriptions if unsubscribe_on_stop is True

        :rtype: None
        :raises RuntimeError: if EventSub is not running
        """
        if not self.__running:
            raise RuntimeError('EventSubWebhook is not running')
        self.logger.debug('shutting down eventsub')
        if self.__hook_runner is not None and self.unsubscribe_on_stop:
            await self.unsubscribe_all_known()
        # ensure all client sessions are closed
        await asyncio.sleep(0.25)
        self._closing = True
        # cleanly shut down the runner
        if self.__hook_runner is not None:
            await self.__hook_runner.shutdown()
            await self.__hook_runner.cleanup()
        self.__hook_runner = None
        self.__running = False
        self.logger.debug('eventsub shut down')

    def _get_transport(self) -> dict:
        return {
            'method': 'webhook',
            'callback': f'{self.callback_url}/callback',
            'secret': self.secret
        }

    async def _build_request_header(self) -> dict:
        token = await self._twitch.get_refreshed_app_token()
        if token is None:
            raise TwitchAuthorizationException('no Authorization set!')
        return {
            'Client-ID': self._twitch.app_id,
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {token}'
        }

    async def _subscribe(self, sub_type: str, sub_version: str, condition: dict, callback, event, is_batching_enabled: Optional[bool] = None) -> str:
        """"Subscribe to Twitch Topic"""
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError('callback needs to be a async function which takes one parameter')
        self.logger.debug(f'subscribe to {sub_type} version {sub_version} with condition {condition}')
        data = {
            'type': sub_type,
            'version': sub_version,
            'condition': condition,
            'transport': self._get_transport()
        }
        if is_batching_enabled is not None:
            data['is_batching_enabled'] = is_batching_enabled

        async with ClientSession(timeout=self._twitch.session_timeout) as session:
            sub_base = self.subscription_url if self.subscription_url is not None else self._twitch.base_url
            r_data = await self._api_post_request(session, sub_base + 'eventsub/subscriptions', data=data)
            result = await r_data.json()
        error = result.get('error')
        if r_data.status == 500:
            raise TwitchBackendException(error)
        if error is not None:
            if error.lower() == 'conflict':
                raise EventSubSubscriptionConflict(result.get('message', ''))
            raise EventSubSubscriptionError(result.get('message'))
        sub_id = result['data'][0]['id']
        self.logger.debug(f'subscription for {sub_type} version {sub_version} with condition {condition} has id {sub_id}')
        self._add_callback(sub_id, callback, event)
        if self.wait_for_subscription_confirm:
            timeout = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=self.wait_for_subscription_confirm_timeout)
            while timeout >= datetime.datetime.utcnow():
                if self._callbacks[sub_id]['active']:
                    return sub_id
                await asyncio.sleep(0.01)
            self._callbacks.pop(sub_id, None)
            raise EventSubSubscriptionTimeout()
        return sub_id

    def _target_token(self) -> AuthType:
        return AuthType.APP

    async def _verify_signature(self, request: 'web.Request') -> bool:
        expected = request.headers['Twitch-Eventsub-Message-Signature']
        hmac_message = request.headers['Twitch-Eventsub-Message-Id'] + \
            request.headers['Twitch-Eventsub-Message-Timestamp'] + await request.text()
        sig = 'sha256=' + hmac.new(bytes(self.secret, 'utf-8'),
                                   msg=bytes(hmac_message, 'utf-8'),
                                   digestmod=hashlib.sha256).hexdigest().lower()
        return sig == expected

    # noinspection PyUnusedLocal
    @staticmethod
    async def __handle_default(request: 'web.Request'):
        return web.Response(text="pyTwitchAPI EventSub")

    async def __handle_challenge(self, request: 'web.Request', data: dict):
        self.logger.debug(f'received challenge for subscription {data.get("subscription", {}).get("id")}')
        if not await self._verify_signature(request):
            self.logger.warning('message signature is not matching! Discarding message')
            return web.Response(status=403)
        await self._activate_callback(data.get('subscription', {}).get('id'))
        return web.Response(text=data.get('challenge'))

    async def _handle_revokation(self, data):
        sub_id: str = data.get('subscription', {}).get('id')
        self.logger.debug(f'got revocation of subscription {sub_id} for reason {data.get("subscription").get("status")}')
        if sub_id not in self._callbacks.keys():
            self.logger.warning(f'unknown subscription {sub_id} got revoked. ignore')
            return
        self._callbacks.pop(sub_id)
        if self.revokation_handler is not None and self._callback_loop is not None:
            t = self._callback_loop.create_task(self.revokation_handler(data)) #type: ignore
            t.add_done_callback(self._task_callback)

    async def __handle_callback(self, request: 'web.Request'):
        try:
            data: dict = await request.json()
        except JSONDecodeError:
            self.logger.error('got request with malformed body! Discarding message')
            return web.Response(status=400)
        if data.get('challenge') is not None:
            return await self.__handle_challenge(request, data)
        sub_id = data.get('subscription', {}).get('id')
        callback = self._callbacks.get(sub_id)
        if callback is None:
            self.logger.error(f'received event for unknown subscription with ID {sub_id}')
        else:
            if not await self._verify_signature(request):
                self.logger.warning('message signature is not matching! Discarding message')
                return web.Response(status=403)
            msg_type = request.headers['Twitch-Eventsub-Message-Type']
            if msg_type.lower() == 'revocation':
                await self._handle_revokation(data)
            else:
                msg_id = request.headers.get('Twitch-Eventsub-Message-Id')
                if msg_id is not None and msg_id in self._msg_id_history:
                    self.logger.warning(f'got message with duplicate id {msg_id}! Discarding message')
                else:
                    self._msg_id_history.append(msg_id)
                    data['metadata'] = {
                        'message_id': msg_id,
                        'message_type': msg_type,
                        'message_timestamp': request.headers['Twitch-Eventsub-Message-Timestamp'],
                        'subscription_type': request.headers['Twitch-Eventsub-Subscription-Type'],
                        'subscription_version': request.headers['Twitch-Eventsub-Subscription-Version'],
                    }
                    dat = callback['event'](**data)
                    if self._callback_loop is not None:
                        t = self._callback_loop.create_task(callback['callback'](dat))
                        t.add_done_callback(self._task_callback)
        return web.Response(status=200)



#  Copyright (c) 2023. Lena "Teekeks" During <info@teawork.de>
"""
EventSub Websocket
------------------

.. note:: EventSub Websocket is targeted at programs which have to subscribe to topics for just a single broadcaster.\n
    Should you need to target multiple broadcasters or are building a server side project, look at :doc:`/modules/twitchAPI.eventsub.webhook`

EventSub lets you listen for events that happen on Twitch.

The EventSub client runs in its own thread, calling the given callback function whenever an event happens.

*******************
Listening to topics
*******************

After you started your EventSub client, you can use the :code:`listen_` prefixed functions to listen to the topics you are interested in.

Look at :ref:`eventsub-available-topics` to find the topics you are interested in.

The function you hand in as callback will be called whenever that event happens with the event data as a parameter,
the type of that parameter is also listed in the link above.

************
Code Example
************

.. code-block:: python

    from twitchAPI.helper import first
    from twitchAPI.twitch import Twitch
    from twitchAPI.oauth import UserAuthenticationStorageHelper
    from twitchAPI.object.eventsub import ChannelFollowEvent
    from twitchAPI.eventsub.websocket import EventSubWebsocket
    from twitchAPI.type import AuthScope
    import asyncio

    APP_ID = 'your_app_id'
    APP_SECRET = 'your_app_secret'
    TARGET_SCOPES = [AuthScope.MODERATOR_READ_FOLLOWERS]


    async def on_follow(data: ChannelFollowEvent):
        # our event happened, lets do things with the data we got!
        print(f'{data.event.user_name} now follows {data.event.broadcaster_user_name}!')


    async def run():
        # create the api instance and get user auth either from storage or website
        twitch = await Twitch(APP_ID, APP_SECRET)
        helper = UserAuthenticationStorageHelper(twitch, TARGET_SCOPES)
        await helper.bind()

        # get the currently logged in user
        user = await first(twitch.get_users())

        # create eventsub websocket instance and start the client.
        eventsub = EventSubWebsocket(twitch)
        eventsub.start()
        # subscribing to the desired eventsub hook for our user
        # the given function (in this example on_follow) will be called every time this event is triggered
        # the broadcaster is a moderator in their own channel by default so specifying both as the same works in this example
        # We have to subscribe to the first topic within 10 seconds of eventsub.start() to not be disconnected.
        await eventsub.listen_channel_follow_v2(user.id, user.id, on_follow)

        # eventsub will run in its own process
        # so lets just wait for user input before shutting it all down again
        try:
            input('press Enter to shut down...')
        except KeyboardInterrupt:
            pass
        finally:
            # stopping both eventsub as well as gracefully closing the connection to the API
            await eventsub.stop()
            await twitch.close()


    asyncio.run(run())
"""
import asyncio
import datetime
import json
import threading
from asyncio import CancelledError
from dataclasses import dataclass
from functools import partial
from time import sleep
from typing import Optional, List, Dict, Callable, Awaitable

import aiohttp
from aiohttp import ClientSession, WSMessage, ClientWebSocketResponse
from collections import deque

from .base import EventSubBase


__all__ = ['EventSubWebsocket']

from twitchAPI.twitch import Twitch
from ..helper import TWITCH_EVENT_SUB_WEBSOCKET_URL, done_task_callback
from ..type import AuthType, UnauthorizedException, TwitchBackendException, EventSubSubscriptionConflict, EventSubSubscriptionError, \
    TwitchAuthorizationException


@dataclass
class Session:
    id: str
    keepalive_timeout_seconds: int
    status: str
    reconnect_url: str

    @classmethod
    def from_twitch(cls, data: dict):
        return cls(
            id=data.get('id'),
            keepalive_timeout_seconds=data.get('keepalive_timeout_seconds'),
            status=data.get('status'),
            reconnect_url=data.get('reconnect_url'),
        )


@dataclass
class Reconnect:
    session: Session
    connection: ClientWebSocketResponse


class EventSubWebsocket(EventSubBase):
    _reconnect: Optional[Reconnect] = None
    
    def __init__(self,
                 twitch: Twitch,
                 connection_url: Optional[str] = None,
                 subscription_url: Optional[str] = None,
                 callback_loop: Optional[asyncio.AbstractEventLoop] = None,
                 revocation_handler: Optional[Callable[[dict], Awaitable[None]]] = None,
                 message_deduplication_history_length: int = 50):
        """
        :param twitch: The Twitch instance to be used
        :param connection_url: Alternative connection URL, useful for development with the twitch-cli
        :param subscription_url: Alternative subscription URL, useful for development with the twitch-cli
        :param callback_loop: The asyncio eventloop to be used for callbacks. \n
            Set this if you or a library you use cares about which asyncio event loop is running the callbacks.
            Defaults to the one used by EventSub Websocket.
        :param revocation_handler: Optional handler for when subscriptions get revoked. |default| :code:`None`
        :param message_deduplication_history_length: The amount of messages being considered for the duplicate message deduplication. |default| :code:`50`
        """
        super().__init__(twitch, 'twitchAPI.eventsub.websocket')
        self.subscription_url: Optional[str] = subscription_url
        """The URL where subscriptions are being sent to. Defaults to :const:`~twitchAPI.helper.TWITCH_API_BASE_URL`"""
        if self.subscription_url is not None and self.subscription_url[-1] != '/':
            self.subscription_url += '/'
        self.connection_url: str = connection_url if connection_url is not None else TWITCH_EVENT_SUB_WEBSOCKET_URL
        """The URL where the websocket connects to. Defaults to :const:`~twitchAPI.helper.TWITCH_EVENT_SUB_WEBSOCKET_URL`"""
        self.active_session: Optional[Session] = None
        """The currently used session"""
        self._running: bool = False
        self._socket_thread = None
        self._startup_complete: bool = False
        self._socket_loop = None
        self._ready: bool = False
        self._closing: bool = False
        self._connection = None
        self._session = None
        self._callback_loop = callback_loop
        self._is_reconnecting: bool = False
        self._active_subscriptions = {}
        self._msg_id_history: deque = deque(maxlen=message_deduplication_history_length)
        self.revokation_handler: Optional[Callable[[dict], Awaitable[None]]] = revocation_handler
        """Optional handler for when subscriptions get revoked."""
        self._task_callback = partial(done_task_callback, self.logger)
        self._reconnect_timeout: Optional[datetime.datetime] = None
        self.reconnect_delay_steps: List[int] = [0, 1, 2, 4, 8, 16, 32, 64, 128]
        """Time in seconds between reconnect attempts"""

    def start(self):
        """Starts the EventSub client

        :raises RuntimeError: If EventSub is already running
        :raises ~twitchAPI.type.UnauthorizedException: If Twitch instance is missing user authentication
        """
        self.logger.debug('starting websocket EventSub...')
        if self._running:
            raise RuntimeError('EventSubWebsocket is already started!')
        if not self._twitch.has_required_auth(AuthType.USER, []):
            raise UnauthorizedException('Twitch needs user authentication')
        self._startup_complete = False
        self._ready = False
        self._closing = False
        self._socket_thread = threading.Thread(target=self._run_socket)
        self._running = True
        self._active_subscriptions = {}
        self._socket_thread.start()
        while not self._startup_complete:
            sleep(0.01)
        self.logger.debug('EventSubWebsocket started up!')

    async def stop(self):
        """Stops the EventSub client

        :raises RuntimeError: If EventSub is not running
        """
        if not self._running:
            raise RuntimeError('EventSubWebsocket is not running')
        self.logger.debug('stopping websocket EventSub...')
        self._startup_complete = False
        self._running = False
        self._ready = False
        if self._socket_loop is not None:
            f = asyncio.run_coroutine_threadsafe(self._stop(), self._socket_loop)
            f.result()

    def _get_transport(self) -> dict:
        return {
            'method': 'websocket',
            'session_id': self.active_session.id
        }

    async def _subscribe(self, sub_type: str, sub_version: str, condition: dict, callback, event, is_batching_enabled: Optional[bool] = None) -> str:
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError('callback needs to be a async function which takes one parameter')
        self.logger.debug(f'subscribe to {sub_type} version {sub_version} with condition {condition}')
        data = {
            'type': sub_type,
            'version': sub_version,
            'condition': condition,
            'transport': self._get_transport()
        }
        if is_batching_enabled is not None:
            data['is_batching_enabled'] = is_batching_enabled
        async with ClientSession(timeout=self._twitch.session_timeout) as session:
            sub_base = self.subscription_url if self.subscription_url is not None else self._twitch.base_url
            r_data = await self._api_post_request(session, sub_base + 'eventsub/subscriptions', data=data)
            result = await r_data.json()
        error = result.get('error')
        if r_data.status == 500:
            raise TwitchBackendException(error)
        if error is not None:
            if error.lower() == 'conflict':
                raise EventSubSubscriptionConflict(result.get('message', ''))
            raise EventSubSubscriptionError(result.get('message'))
        sub_id = result['data'][0]['id']
        self.logger.debug(f'subscription for {sub_type} version {sub_version} with condition {condition} has id {sub_id}')
        self._add_callback(sub_id, callback, event)
        self._callbacks[sub_id]['active'] = True
        self._active_subscriptions[sub_id] = {
            'sub_type': sub_type,
            'sub_version': sub_version,
            'condition': condition,
            'callback': callback,
            'event': event
        }
        return sub_id

    def _target_token(self) -> AuthType:
        return AuthType.USER

    async def _connect(self, is_startup: bool = False):
        if is_startup:
            self.logger.debug(f'connecting to {self.connection_url}...')
        else:
            self._is_reconnecting = True
            self.logger.debug(f'reconnecting using {self.connection_url}...')
        self._reconnect_timeout = None
        if self._connection is not None and not self._connection.closed:
            await self._connection.close()
            while not self._connection.closed:
                await asyncio.sleep(0.1)
        retry = 0
        need_retry = True
        if self._session is None:
            self._session = aiohttp.ClientSession(timeout=self._twitch.session_timeout)
        while need_retry and retry < len(self.reconnect_delay_steps):
            need_retry = False
            try:
                self._connection = await self._session.ws_connect(self.connection_url)
            except Exception:
                self.logger.warning(f'connection attempt failed, retry in {self.reconnect_delay_steps[retry]} seconds...')
                await asyncio.sleep(self.reconnect_delay_steps[retry])
                retry += 1
                need_retry = True
        if retry >= len(self.reconnect_delay_steps):
            raise TwitchBackendException(f'can\'t connect to EventSub websocket {self.connection_url}')

    def _run_socket(self):
        self._socket_loop = asyncio.new_event_loop()
        if self._callback_loop is None:
            self._callback_loop = self._socket_loop
        asyncio.set_event_loop(self._socket_loop)

        self._socket_loop.run_until_complete(self._connect(is_startup=True))

        self._tasks = [
            asyncio.ensure_future(self._task_receive(), loop=self._socket_loop),
            asyncio.ensure_future(self._task_reconnect_handler(), loop=self._socket_loop)
        ]
        self._socket_loop.run_until_complete(self._keep_loop_alive())

    async def _stop(self):
        await self._connection.close()
        await self._session.close()
        await asyncio.sleep(0.25)
        self._connection = None
        self._session = None
        self._closing = True

    async def _keep_loop_alive(self):
        while not self._closing:
            await asyncio.sleep(0.1)

    async def _task_reconnect_handler(self):
        try:
            while not self._closing:
                await asyncio.sleep(0.1)
                if self._reconnect_timeout is None:
                    continue
                if self._reconnect_timeout <= datetime.datetime.now():
                    self.logger.warning('keepalive missed, connection lost. reconnecting...')
                    self._reconnect_timeout = None
                    await self._connect(is_startup=False)
        except CancelledError:
            return

    async def _task_receive(self):
        handler: Dict[str, Callable] = {
            'session_welcome': self._handle_welcome,
            'session_keepalive': self._handle_keepalive,
            'notification': self._handle_notification,
            'session_reconnect': self._handle_reconnect,
            'revocation': self._handle_revocation
        }
        try:
            while not self._closing:
                if self._connection.closed:
                    await asyncio.sleep(0.01)
                    continue
                message: WSMessage = await self._connection.receive()
                if message.type == aiohttp.WSMsgType.TEXT:
                    data = json.loads(message.data)
                    _type = data.get('metadata', {}).get('message_type')
                    _handler = handler.get(_type)
                    if _handler is not None:
                        asyncio.ensure_future(_handler(data))
                    # debug
                    else:
                        self.logger.warning(f'got message for unknown message_type: {_type}, ignoring...')
                elif message.type == aiohttp.WSMsgType.CLOSE:
                    msg_lookup = {
                        4000: "4000 - Internal server error",
                        4001: "4001 - Client sent inbound traffic",
                        4002: "4002 - Client failed ping-pong",
                        4003: "4003 - Connection unused, you have to create a subscription within 10 seconds",
                        4004: "4004 - Reconnect grace time expired",
                        4005: "4005 - Network timeout",
                        4006: "4006 - Network error",
                        4007: "4007 - Invalid reconnect"
                    }
                    self.logger.info(f'Websocket closing: {msg_lookup.get(message.data, f" {message.data} - Unknown")}')
                elif message.type == aiohttp.WSMsgType.CLOSING:
                    if self._reconnect and self._reconnect.session.status == "connected":
                        self._connection = self._reconnect.connection
                        self.active_session = self._reconnect.session
                        self._reconnect = None
                        self.logger.debug("websocket session_reconnect completed")
                        continue
                elif message.type == aiohttp.WSMsgType.CLOSED:
                    self.logger.debug('websocket is closing')
                    if self._running:
                        if self._is_reconnecting:
                            continue
                        try:
                            await self._connect(is_startup=False)
                        except TwitchBackendException:
                            self.logger.exception('Connection to EventSub websocket lost and unable to reestablish connection!')
                            break
                    else:
                        break
                elif message.type == aiohttp.WSMsgType.ERROR:
                    self.logger.warning('error in websocket: ' + str(self._connection.exception()))
                    break
        except CancelledError:
            return

    async def _build_request_header(self):
        token = await self._twitch.get_refreshed_user_auth_token()
        if token is None:
            raise TwitchAuthorizationException('no Authorization set!')
        return {
            'Client-ID': self._twitch.app_id,
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {token}'
        }

    async def _unsubscribe_hook(self, topic_id: str) -> bool:
        self._active_subscriptions.pop(topic_id, None)
        return True

    async def _resubscribe(self):
        self.logger.debug('resubscribe to all active subscriptions of this websocket...')
        subs = self._active_subscriptions.copy()
        self._active_subscriptions = {}
        try:
            for sub in subs.values():
                await self._subscribe(**sub)
        except BaseException:
            self.logger.exception('exception while resubscribing')
            if not self._active_subscriptions:  # Restore old subscriptions for next reconnect
                self._active_subscriptions = subs
            return
        self.logger.debug('done resubscribing!')

    def _reset_timeout(self):
        self._reconnect_timeout = datetime.datetime.now() + datetime.timedelta(seconds=self.active_session.keepalive_timeout_seconds*2)

    async def _handle_revocation(self, data: dict):
        _payload = data.get('payload', {})
        sub_id: str = _payload.get('subscription', {}).get('id')
        self.logger.debug(f'got revocation of subscription {sub_id} for reason {_payload.get("subscription").get("status")}')
        if sub_id not in self._active_subscriptions.keys():
            self.logger.warning(f'unknown subscription {sub_id} got revoked. ignore')
            return
        self._active_subscriptions.pop(sub_id)
        self._callbacks.pop(sub_id)
        if self.revokation_handler is not None:
            t = self._callback_loop.create_task(self.revokation_handler(_payload))
            t.add_done_callback(self._task_callback)

    async def _handle_reconnect(self, data: dict):
        session = data.get('payload', {}).get('session', {})
        new_session = Session.from_twitch(session)
        self.logger.debug(f"got request from websocket to reconnect, reconnect url: {new_session.reconnect_url}")
        self._reset_timeout()
        new_connection = None
        retry = 0
        need_retry = True
        while need_retry and retry <= 5:  # We only have up to 30 seconds to move to new connection
            need_retry = False
            try:
                new_connection = await self._session.ws_connect(new_session.reconnect_url)
            except Exception as err:
                self.logger.warning(f"reconnection attempt failed because {err}, retry in {self.reconnect_delay_steps[retry]} seconds...")
                await asyncio.sleep(self.reconnect_delay_steps[retry])
                retry += 1
                need_retry = True
        if new_connection is None:  # We failed to establish new connection, do nothing and force a full refresh
            self.logger.warning(f"Failed to establish connection to {new_session.reconnect_url}, Twitch will close and we'll reconnect")
            return
        reconnect = Reconnect(session=new_session, connection=new_connection)
        try:
            message: WSMessage = await reconnect.connection.receive(timeout=30)
        except asyncio.TimeoutError:
            await reconnect.connection.close()
            self.logger.warning(f"Reconnect socket got a timeout waiting for first message {reconnect.session}")
            return
        self._reset_timeout()
        if message.type != aiohttp.WSMsgType.TEXT:
            self.logger.warning(f"Reconnect socket got an unknown message {message}")
            await reconnect.connection.close()
            return
        data = message.json()
        message_type = data.get('metadata', {}).get('message_type')
        if message_type != "session_welcome":
            self.logger.warning(f"Reconnect socket got a non session_welcome first message {data}")
            await reconnect.connection.close()
            return
        session_dict = data.get('payload', {}).get('session', {})
        reconnect.session = Session.from_twitch(session_dict)
        self._reconnect = reconnect
        await self._connection.close()  # This will wake up _task_receive with a CLOSING message

    async def _handle_welcome(self, data: dict):
        session = data.get('payload', {}).get('session', {})
        self.active_session = Session.from_twitch(session)
        self.logger.debug(f'new session id: {self.active_session.id}')
        self._reset_timeout()
        if self._is_reconnecting:
            await self._resubscribe()
        self._is_reconnecting = False
        self._startup_complete = True

    async def _handle_keepalive(self, data: dict):
        self.logger.debug('got session keep alive')
        self._reset_timeout()

    async def _handle_notification(self, data: dict):
        self._reset_timeout()
        _payload = data.get('payload', {})
        _payload['metadata'] = data.get('metadata', {})
        sub_id = _payload.get('subscription', {}).get('id')
        callback = self._callbacks.get(sub_id)
        if callback is None:
            self.logger.error(f'received event for unknown subscription with ID {sub_id}')
        else:
            msg_id = _payload['metadata'].get('message_id')
            if msg_id is not None and msg_id in self._msg_id_history:
                self.logger.warning(f'got message with duplicate id {msg_id}! Discarding message')
            else:
                t = self._callback_loop.create_task(callback['callback'](callback['event'](**_payload)))
                t.add_done_callback(self._task_callback)

